@@ -19,7 +19,7 @@ import click
 import yaml
 
 import vault_cli
-from vault_cli import client, environment, exceptions, settings, ssh, types
+from vault_cli import client, environment, exceptions, settings, ssh, types, utils
 
 logger = logging.getLogger(__name__)
 
@@ -59,7 +59,7 @@ def handle_errors():
     try:
         yield
     except exceptions.VaultException as exc:
-        raise click.ClickException(str(exc))
+        raise click.ClickException("\n".join(utils.extract_error_messages(exc)))
 
 
 def print_version(ctx, __, value):
@@ -132,7 +132,7 @@ def repr_octal(value: Optional[int]) -> Optional[str]:
     "--safe-write/--unsafe-write",
     default=settings.DEFAULTS.safe_write,
     help="When activated, you can't overwrite a secret without "
-    'passing "--force" (in commands "set", "mv", etc)',
+    'passing "--force" (in commands "set", "mv", "cp", etc)',
 )
 @click.option(
     "--render/--no-render",
@@ -140,10 +140,15 @@ def repr_octal(value: Optional[int]) -> Optional[str]:
     help="Render templated values",
 )
 @click.option(
-    "--umask", callback=click_octal, help="Set umask for newly created files.",
+    "--umask",
+    callback=click_octal,
+    help="Set umask for newly created files.",
 )
 @click.option(
-    "-v", "--verbose", count=True, help="Use multiple times to increase verbosity",
+    "-v",
+    "--verbose",
+    count=True,
+    help="Use multiple times to increase verbosity",
 )
 @click.option(
     "--config-file",
@@ -405,10 +410,15 @@ def delete(client_obj: client.VaultClientBase, name: str, key: Optional[str]) ->
 @click.option(
     "-o",
     "--omit-single-key/--no-omit-single-key",
-    is_flag=True,
     default=False,
     help="When the secret has only one key, don't use that key to build the name of the environment variable",
 )
+@click.option(
+    "-f",
+    "--force/--no-force",
+    default=False,
+    help="Run the command even if there is a problem while reading secrets",
+)
 @click.argument("command", nargs=-1, required=True)
 @click.pass_obj
 @handle_errors()
@@ -416,6 +426,7 @@ def env(
     client_obj: client.VaultClientBase,
     path: Sequence[str],
     omit_single_key: bool,
+    force: bool,
     command: Sequence[str],
 ) -> NoReturn:
     """
@@ -450,6 +461,7 @@ def env(
         path_with_key, _, prefix = path.partition("=")
         path, _, filter_key = path_with_key.partition(":")
 
+        env_updates = {}
         env_updates = environment.get_envvars(
             vault_client=client_obj,
             path=path,
@@ -457,15 +469,20 @@ def env(
             omit_single_key=omit_single_key,
             filter_key=filter_key,
         )
+
         env_secrets.update(env_updates)
 
+    if bool(client_obj.errors) and not force:
+        raise click.ClickException("There was an error while reading the secrets.")
     environment.exec_command(command=command, environment=env_secrets)
 
 
 @cli.command("dump-config")
 @click.pass_obj
 @handle_errors()
-def dump_config(client_obj: client.VaultClientBase,) -> None:
+def dump_config(
+    client_obj: client.VaultClientBase,
+) -> None:
     """
     Display settings in the format of a config file.
     """
@@ -534,6 +551,38 @@ def mv(
         raise click.ClickException(str(exc))
 
 
+@cli.command()
+@click.argument("source", required=True)
+@click.argument("dest", required=True)
+@click.option(
+    "--force/--no-force",
+    "-f",
+    is_flag=True,
+    default=None,
+    help="In case the path already holds a secret, allow overwriting it "
+    "(this is necessary only if --safe-write is set).",
+)
+@click.pass_obj
+@handle_errors()
+def cp(
+    client_obj: client.VaultClientBase, source: str, dest: str, force: Optional[bool]
+) -> None:
+    """
+    Recursively copy secrets from source to destination path.
+    """
+    try:
+        for old_path, new_path in client_obj.copy_secrets(
+            source=source, dest=dest, force=force, generator=True
+        ):
+            click.echo(f"Copy '{old_path}' to '{new_path}'")
+    except exceptions.VaultOverwriteSecretError as exc:
+        raise click.ClickException(
+            f"Secret already exists at {exc.path}. Use -f to force overwriting."
+        )
+    except exceptions.VaultMixSecretAndFolder as exc:
+        raise click.ClickException(str(exc))
+
+
 @cli.command()
 @click.argument(
     "template",
@@ -625,20 +674,30 @@ def ssh_(
     if ":" not in key:
         raise click.UsageError("Format for private_key: `path/to/private_key:key`")
     private_key_name, private_key_key = key.rsplit(":", 1)
-    private_key_secret = client_obj.get_secret(private_key_name, private_key_key)
+    private_key_secret_obj = client_obj.get_secret(private_key_name, private_key_key)
+    private_key_secret = ensure_str(secret=private_key_secret_obj, path=key)
 
     passphrase_secret = None
     if passphrase:
         if ":" not in passphrase:
             raise click.UsageError("Format for passphrase: `path/to/passphrase:key`")
         passphrase_name, passphrase_key = passphrase.rsplit(":", 1)
-        passphrase_secret = client_obj.get_secret(passphrase_name, passphrase_key)
+        passphrase_secret_obj = client_obj.get_secret(passphrase_name, passphrase_key)
+        passphrase_secret = ensure_str(secret=passphrase_secret_obj, path=passphrase)
 
     ssh.add_key(key=private_key_secret, passphrase=passphrase_secret)
 
     environment.exec_command(command=command)
 
 
+def ensure_str(secret, path) -> str:
+    if not isinstance(secret, str):
+        raise exceptions.VaultWrongType(
+            f"secret at {path} is not a string but {type(secret)}"
+        )
+    return secret
+
+
 # This is purposedly not called as a click subcommand, and we cannot take any argument
 def askpass():
     print(os.environ[ssh.SSH_PASSPHRASE_ENVVAR])
